#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json, numpy as np

def wrap_outputs_to_json(raw_outputs: dict = {"keys": [], "outputs": {}}) -> str:
    """helper function to wrap outputs into a pure python object
    and employ json dumps on it.
    """
    assert ( raw_outputs.get("keys", None) )
    assert ( raw_outputs.get("outputs", None) )

    keys = raw_outputs["keys"]
    pythonized_outputs = {}
    for k, o in raw_outputs["outputs"].items(): 
        if o is None: return None
        if isinstance(o, np.ndarray): o = o.tolist()
        pythonized_outputs[k] = o

    outputs = json.dumps({
            "keys": keys, 
            "outputs": pythonized_outputs
        }, ensure_ascii=False)
    return outputs




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

class Sum(object):
    """sum a numpy arrray.
    """
    def __init__(self, default=np.nan):
        self.default = default

    def __call__(self, np_arr):
        try:
            return np.sum(np_arr) if np_arr.shape[0] != 0 else self.default
        except:
            raise Exception("Sum encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))

class Mean(object):
    """get mean of a numpy arrray.
    """
    def __init__(self, default=np.nan):
        self.default = default

    def __call__(self, np_arr):
        try:
            return np.mean(np_arr) if np_arr.shape[0] != 0 else self.default
        except:
            raise Exception("Mean encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))

class Std(object):
    """get mean of a numpy arrray.
    """
    def __init__(self, default=np.nan):
        self.default = default

    def __call__(self, np_arr):
        try:
            return np.std(np_arr) if np_arr.shape[0] != 0 else self.default
        except:
            raise Exception("Std encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))

class Max(object):
    """get max of a numpy arrray.
    """
    def __init__(self, default=np.nan):
        self.default = default

    def __call__(self, np_arr):
        try:
            return np.max(np_arr) if np_arr.shape[0] != 0 else self.default
        except:
            raise Exception("Max encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))

class Min(object):
    """get max of a numpy arrray.
    """
    def __init__(self, default=np.nan):
        self.default = default

    def __call__(self, np_arr):
        try:
            return np.min(np_arr) if np_arr.shape[0] != 0 else self.default
        except:
            raise Exception("Min encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))

class Quantile(object):
    """get max of a numpy arrray.
    """
    def __init__(self, q, interpolation='linear', default=np.nan):
        self.q = q
        self.interpolation = interpolation
        self.default = default

    def __call__(self, np_arr):
        try:
            return np.percentile(np_arr, self.q, interpolation=self.interpolation) if np_arr.shape[0] != 0 else self.default
        except:
            raise Exception("Quantile encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys, itertools, json, numpy as np

def get_tsv_headers(path: str) -> list:
    """helper function to get tsv file headers.
    """
    fd = open(path, "r")
    headers = fd.readline().replace("\n", "").split("\t")
    fd.close()
    return headers

class TSVPlaceholder(object):
    def __init__(self, names: list = [], name2index: dict = {}) -> None:
        """initialize a placeholder handler for tsv file. the params passed in are usually ignored.
        unless the file is broken, you may want to manually define these.
        """
        self.names = names
        self.name2index = name2index

    @classmethod
    def init_from_file(cls, path: str):
        """initialize the settings from file for later forward use.
        """
        names = get_tsv_headers(path)
        name2index= {n: i for i, n in enumerate(names)}
        return cls(names, name2index)

    def __call__(self, chunk: list):
        """forward the data into the placeholders.
        """
        assert (self.names)
        if self.names[0] in chunk[0]: return

        chunk = np.array( [entry.replace("\n", "").split("\t") for entry in chunk], dtype=object )

        source = {}
        for name in self.names:
            index = self.name2index.get(name, -1)
            if (index == -1): raise Exception("Wrong name2index mapping.")
            data = chunk[:, index].reshape(-1, 1)
            source[name] = data
        return source


class TSVDataLoader(object):
    def __init__(self, groupby: list, names: list = [], name2index: dict = {}, fd: int = None, hard_chunk_size: int = 1024000, auto_wrangle: bool = True) -> None:
        """initialize an itertools groupby handler for tsv file. the params passed in are usually ignored.
        unless the file is broken, you may want to manually define these.
        """
        self.groupby = groupby
        self.names = names
        self.name2index = name2index
        self.hard_chunk_size = hard_chunk_size
        self.auto_wrangle = auto_wrangle
        self.fd = fd

    @classmethod
    def init_from_file(cls, groupby: list, path: str, is_fdin: bool = True, encoding: str = "utf-8", hard_chunk_size: int = 1024000, auto_wrangle: bool = True):
        """initialize the settings from file for later forward use.
        """
        names = get_tsv_headers(path)
        name2index= {n: i for i, n in enumerate(names)}
        fd = open(path, "r", encoding=encoding) if is_fdin else None
        return cls(groupby, names, name2index, fd, hard_chunk_size, auto_wrangle)

    def _auto_wrangle(self, line):
        words = []
        quote = False
        for char in line:
            if char == '"':
                quote = not quote
            elif not quote and char == '\t':
                yield ''.join(words)
                words = []
            else:
                words.append(char)
        yield ''.join(words)

    def _groupby_func(self, line):
        """helper function as a groupby function.
        """
        indexes = [self.name2index.get(k, -1) for k in self.groupby]
        if (-1 in indexes): raise Exception("Wrong groupby keys.")

        line = line.replace("\n", "")
        if self.auto_wrangle:
            line = list( self._auto_wrangle(line) )
        else:
            line = line.split("\t")
        return [line[i] for i in indexes]

    def __call__(self, fdin: int = None):
        """build a data generator.
        """
        self.fd = fdin if fdin else self.fd
        assert (self.fd)
        for k, g in itertools.groupby(self.fd, key=self._groupby_func):
            chunk = []
            should_yield = True
            for i, e in enumerate(g):
                if i >= self.hard_chunk_size: 
                    print("Warning: The chunk exceeds the hard size. So skips to continue. Its key is %s" % (str(k)))
                    should_yield = False
                    break
                chunk.append(e)
            if should_yield: yield k, chunk




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

from dateutil.parser import parse

class WrangleFloat(object):
    """transform every element within a numpy array into float object.
    """
    def __init__(self, default=np.nan):
        self.default = default
        self._func = np.vectorize(self._wrangle)

    def _wrangle(self, x):
        if x == "":
            return self.default
        else:
            return float(x)

    def __call__(self, np_arr):
        try:
            return self._func(np_arr) if np_arr.shape[0] != 0 else np_arr
        except:
            raise Exception("WrangleFloat encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))


class WrangleStr(object):
    """transform every element within a numpy array into string object.
    """
    def __init__(self, default=np.nan):
        self.default = default
        self._func = np.vectorize(self._wrangle)

    def _wrangle(self, x):
        if x == "":
            return self.default
        else:
            return str(x)

    def __call__(self, np_arr):
        try:
            return self._func(np_arr) if np_arr.shape[0] != 0 else np_arr
        except:
            raise Exception("WrangleStr encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))

class WrangleNormalDatetime(object):
    """transform every element within a numpy array into datetime object.
    """
    def __init__(self, default=np.nan):
        self.default = default
        self._func = np.vectorize(self._wrangle)

    def _wrangle(self, x):
        if x == "":
            return self.default
        else:
            return parse(str(x))

    def __call__(self, np_arr):
        try:
            return self._func(np_arr) if np_arr.shape[0] != 0 else np_arr
        except:
            raise Exception("WrangleNormalDatetime encounters unexpected error need hot-fixing, the data you are dealing with is: \n{0}\n".format(np_arr))
